@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/optimizer/util/clauses.c,v 1.154.2.4 2005/04/14 21:44:35 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/optimizer/util/clauses.c,v 1.154.2.7 2007/05/01 18:54:24 tgl Exp $
  *
  * HISTORY
  *	  AUTHOR			DATE			MAJOR EVENT
@@ -31,6 +31,7 @@
 #include "optimizer/var.h"
 #include "parser/analyze.h"
 #include "parser/parse_clause.h"
+#include "parser/parse_coerce.h"
 #include "parser/parse_expr.h"
 #include "tcop/tcopprot.h"
 #include "utils/acl.h"
@@ -1414,6 +1415,7 @@ eval_const_expressions_mutator(Node *node, List *active_fns)
 			newrelabel->arg = (Expr *) arg;
 			newrelabel->resulttype = relabel->resulttype;
 			newrelabel->resulttypmod = relabel->resulttypmod;
+			newrelabel->relabelformat = relabel->relabelformat;
 			return (Node *) newrelabel;
 		}
 	}
@@ -1758,7 +1760,6 @@ inline_function(Oid funcid, Oid result_type, List *args,
 {
 	Form_pg_proc funcform = (Form_pg_proc) GETSTRUCT(func_tuple);
 	char		result_typtype;
-	bool		polymorphic = false;
 	Oid			argtypes[FUNC_MAX_ARGS];
 	char	   *src;
 	Datum		tmp;
@@ -1792,10 +1793,8 @@ inline_function(Oid funcid, Oid result_type, List *args,
 	if (result_typtype != 'b' &&
 		result_typtype != 'd')
 	{
-		if (funcform->prorettype == ANYARRAYOID ||
-			funcform->prorettype == ANYELEMENTOID)
-			polymorphic = true;
-		else
+		if (funcform->prorettype != ANYARRAYOID &&
+			funcform->prorettype != ANYELEMENTOID)
 			return NULL;
 	}
 
@@ -1814,7 +1813,6 @@ inline_function(Oid funcid, Oid result_type, List *args,
 		if (argtypes[i] == ANYARRAYOID ||
 			argtypes[i] == ANYELEMENTOID)
 		{
-			polymorphic = true;
 			argtypes[i] = exprType((Node *) nth(i, args));
 		}
 	}
@@ -1891,16 +1889,14 @@ inline_function(Oid funcid, Oid result_type, List *args,
 	newexpr = (Node *) ((TargetEntry *) lfirst(querytree->targetList))->expr;
 
 	/*
-	 * If the function has any arguments declared as polymorphic types,
-	 * then it wasn't type-checked at definition time; must do so now.
-	 * (This will raise an error if wrong, but that's okay since the
-	 * function would fail at runtime anyway.  Note we do not try this
-	 * until we have verified that no rewriting was needed; that's
-	 * probably not important, but let's be careful.)
+	 * Make sure the function (still) returns what it's declared to.  This will
+	 * raise an error if wrong, but that's okay since the function would fail
+	 * at runtime anyway.  Note we do not try this until we have verified that
+	 * no rewriting was needed; that's probably not important, but let's be
+	 * careful.
 	 */
-	if (polymorphic)
-		check_sql_fn_retval(result_type, get_typtype(result_type),
-							querytree_list);
+	check_sql_fn_retval(result_type, get_typtype(result_type),
+						querytree_list);
 
 	/*
 	 * Additional validity checks on the expression.  It mustn't return a
@@ -1985,6 +1981,21 @@ inline_function(Oid funcid, Oid result_type, List *args,
 
 	MemoryContextDelete(mycxt);
 
+	/*
+	 * Since check_sql_fn_retval allows binary-compatibility cases, the
+	 * expression we now have might return some type that's only binary
+	 * compatible with the original expression result type.  To avoid
+	 * confusing matters, insert a RelabelType in such cases.
+	 */
+	if (exprType(newexpr) != result_type)
+	{
+		Assert(IsBinaryCoercible(exprType(newexpr), result_type));
+		newexpr = (Node *) makeRelabelType((Expr *) newexpr,
+										   result_type,
+										   -1,
+										   COERCE_IMPLICIT_CAST);
+	}
+
 	/*
 	 * Recursively try to simplify the modified expression.  Here we must
 	 * add the current function to the context list of active functions.
