@@ -1,17 +1,16 @@
 /*-------------------------------------------------------------------------
+ *
  * nabstime.c
  *	  Utilities for the built-in type "AbsoluteTime".
  *	  Functions for the built-in type "RelativeTime".
  *	  Functions for the built-in type "TimeInterval".
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/nabstime.c,v 1.100.2.4 2003/05/04 04:30:35 tgl Exp $
- *
- * NOTES
+ *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/nabstime.c,v 1.117.2.2 2005/05/26 02:14:31 neilc Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -23,11 +22,8 @@
 #include <float.h>
 #include <limits.h>
 
-#if !(defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE))
-#include <sys/timeb.h>
-#endif
-
 #include "access/xact.h"
+#include "libpq/pqformat.h"
 #include "miscadmin.h"
 #include "utils/builtins.h"
 
@@ -87,79 +83,26 @@ static int istinterval(char *i_string,
 			AbsoluteTime *i_end);
 
 
-/* GetCurrentAbsoluteTime()
- * Get the current system time. Set timezone parameters if not specified elsewhere.
- * Define HasCTZSet to allow clients to specify the default timezone.
+/*
+ * GetCurrentAbsoluteTime()
  *
- * Returns the number of seconds since epoch (January 1 1970 GMT)
+ * Get the current system time (relative to Unix epoch).
  */
 AbsoluteTime
 GetCurrentAbsoluteTime(void)
 {
 	time_t		now;
 
-#if defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
-	struct tm  *tm;
-
 	now = time(NULL);
-#else
-	struct timeb tb;			/* the old V7-ism */
-
-	ftime(&tb);
-	now = tb.time;
-#endif
-
-	if (!HasCTZSet)
-	{
-#if defined(HAVE_TM_ZONE)
-		tm = localtime(&now);
-
-		CTimeZone = -tm->tm_gmtoff;		/* tm_gmtoff is Sun/DEC-ism */
-		CDayLight = (tm->tm_isdst > 0);
-
-#ifdef NOT_USED
-
-		/*
-		 * XXX is there a better way to get local timezone string w/o
-		 * tzname? - tgl 97/03/18
-		 */
-		strftime(CTZName, MAXTZLEN, "%Z", tm);
-#endif
-
-		/*
-		 * XXX FreeBSD man pages indicate that this should work - thomas
-		 * 1998-12-12
-		 */
-		StrNCpy(CTZName, tm->tm_zone, MAXTZLEN+1);
-
-#elif defined(HAVE_INT_TIMEZONE)
-		tm = localtime(&now);
-
-		CDayLight = tm->tm_isdst;
-		CTimeZone = ((tm->tm_isdst > 0) ? (TIMEZONE_GLOBAL - 3600) : TIMEZONE_GLOBAL);
-		StrNCpy(CTZName, tzname[tm->tm_isdst], MAXTZLEN+1);
-#else							/* neither HAVE_TM_ZONE nor
-								 * HAVE_INT_TIMEZONE */
-		CTimeZone = tb.timezone * 60;
-		CDayLight = (tb.dstflag != 0);
-
-		/*
-		 * XXX does this work to get the local timezone string in V7? -
-		 * tgl 97/03/18
-		 */
-		strftime(CTZName, MAXTZLEN, "%Z", localtime(&now));
-#endif
-	}
-
 	return (AbsoluteTime) now;
-}	/* GetCurrentAbsoluteTime() */
+}
 
 
-/* GetCurrentAbsoluteTimeUsec()
- * Get the current system time. Set timezone parameters if not specified elsewhere.
- * Define HasCTZSet to allow clients to specify the default timezone.
+/*
+ * GetCurrentAbsoluteTimeUsec()
  *
- * Returns the number of seconds since epoch (January 1 1970 GMT)
+ * Get the current system time (relative to Unix epoch), including fractional
+ * seconds expressed as microseconds.
  */
 AbsoluteTime
 GetCurrentAbsoluteTimeUsec(int *usec)
@@ -167,85 +110,54 @@ GetCurrentAbsoluteTimeUsec(int *usec)
 	time_t		now;
 	struct timeval tp;
 
-#ifdef NOT_USED
-	struct timezone tpz;
-#endif
-#if defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
-	struct tm  *tm;
-
-#else
-	struct timeb tb;			/* the old V7-ism */
-#endif
-
 	gettimeofday(&tp, NULL);
-
 	now = tp.tv_sec;
 	*usec = tp.tv_usec;
-
-#ifdef NOT_USED
-#if defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
-	now = time(NULL);
-#else
-	ftime(&tb);
-	now = tb.time;
-#endif
-#endif
-
-	if (!HasCTZSet)
-	{
-#if defined(HAVE_TM_ZONE)
-		tm = localtime(&now);
-
-		CTimeZone = -tm->tm_gmtoff;		/* tm_gmtoff is Sun/DEC-ism */
-		CDayLight = (tm->tm_isdst > 0);
-
-#ifdef NOT_USED
-
-		/*
-		 * XXX is there a better way to get local timezone string w/o
-		 * tzname? - tgl 97/03/18
-		 */
-		strftime(CTZName, MAXTZLEN, "%Z", tm);
-#endif
-
-		/*
-		 * XXX FreeBSD man pages indicate that this should work - thomas
-		 * 1998-12-12
-		 */
-		StrNCpy(CTZName, tm->tm_zone, MAXTZLEN+1);
-
-#elif defined(HAVE_INT_TIMEZONE)
-		tm = localtime(&now);
-
-		CDayLight = tm->tm_isdst;
-		CTimeZone = ((tm->tm_isdst > 0) ? (TIMEZONE_GLOBAL - 3600) : TIMEZONE_GLOBAL);
-		StrNCpy(CTZName, tzname[tm->tm_isdst], MAXTZLEN+1);
-#else							/* neither HAVE_TM_ZONE nor
-								 * HAVE_INT_TIMEZONE */
-		CTimeZone = tb.timezone * 60;
-		CDayLight = (tb.dstflag != 0);
-
-		/*
-		 * XXX does this work to get the local timezone string in V7? -
-		 * tgl 97/03/18
-		 */
-		strftime(CTZName, MAXTZLEN, "%Z", localtime(&now));
-#endif
-	};
-
 	return (AbsoluteTime) now;
-}	/* GetCurrentAbsoluteTimeUsec() */
+}
 
 
+/*
+ * AbsoluteTimeUsecToTimestampTz()
+ *
+ * Convert system time including microseconds to TimestampTz representation.
+ */
+TimestampTz
+AbsoluteTimeUsecToTimestampTz(AbsoluteTime sec, int usec)
+{
+	TimestampTz result;
+
+#ifdef HAVE_INT64_TIMESTAMP
+	result = ((sec - ((POSTGRES_EPOCH_JDATE - UNIX_EPOCH_JDATE) * 86400))
+			  * INT64CONST(1000000)) + usec;
+#else
+	result = sec - ((POSTGRES_EPOCH_JDATE - UNIX_EPOCH_JDATE) * 86400)
+		+ (usec / 1000000.0);
+#endif
+
+	return result;
+}
+
+
+/*
+ * GetCurrentDateTime()
+ *
+ * Get the transaction start time ("now()") broken down as a struct tm.
+ */
 void
 GetCurrentDateTime(struct tm * tm)
 {
 	int			tz;
 
 	abstime2tm(GetCurrentTransactionStartTime(), &tz, tm, NULL);
-}	/* GetCurrentDateTime() */
-
+}
 
+/*
+ * GetCurrentTimeUsec()
+ *
+ * Get the transaction start time ("now()") broken down as a struct tm,
+ * including fractional seconds and timezone offset.
+ */
 void
 GetCurrentTimeUsec(struct tm * tm, fsec_t *fsec, int *tzp)
 {
@@ -253,31 +165,23 @@ GetCurrentTimeUsec(struct tm * tm, fsec_t *fsec, int *tzp)
 	int			usec;
 
 	abstime2tm(GetCurrentTransactionStartTimeUsec(&usec), &tz, tm, NULL);
-	/* Note: don't pass NULL tzp directly to abstime2tm */
+	/* Note: don't pass NULL tzp to abstime2tm; affects behavior */
 	if (tzp != NULL)
 		*tzp = tz;
 #ifdef HAVE_INT64_TIMESTAMP
 	*fsec = usec;
 #else
-	*fsec = usec * 1.0e-6;
+	*fsec = usec / 1000000.0;
 #endif
-}	/* GetCurrentTimeUsec() */
+}
 
 
 void
 abstime2tm(AbsoluteTime _time, int *tzp, struct tm * tm, char **tzn)
 {
 	time_t		time = (time_t) _time;
-
-#if defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
 	struct tm  *tx;
 
-#else
-	struct timeb tb;			/* the old V7-ism */
-
-	ftime(&tb);
-#endif
-
 	/*
 	 * If HasCTZSet is true then we have a brute force time zone
 	 * specified. Go ahead and rotate to the local time zone since we will
@@ -286,11 +190,10 @@ abstime2tm(AbsoluteTime _time, int *tzp, struct tm * tm, char **tzn)
 	if (HasCTZSet && (tzp != NULL))
 		time -= CTimeZone;
 
-#if defined(HAVE_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
 	if ((!HasCTZSet) && (tzp != NULL))
-		tx = localtime((time_t *) &time);
+		tx = localtime(&time);
 	else
-		tx = gmtime((time_t *) &time);
+		tx = gmtime(&time);
 
 	tm->tm_year = tx->tm_year + 1900;
 	tm->tm_mon = tx->tm_mon + 1;
@@ -336,7 +239,10 @@ abstime2tm(AbsoluteTime _time, int *tzp, struct tm * tm, char **tzn)
 				 */
 				StrNCpy(*tzn, tm->tm_zone, MAXTZLEN + 1);
 				if (strlen(tm->tm_zone) > MAXTZLEN)
-					elog(WARNING, "Invalid timezone \'%s\'", tm->tm_zone);
+					ereport(WARNING,
+							(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+							 errmsg("invalid time zone name: \"%s\"",
+									tm->tm_zone)));
 			}
 		}
 	}
@@ -369,13 +275,15 @@ abstime2tm(AbsoluteTime _time, int *tzp, struct tm * tm, char **tzn)
 				 */
 				StrNCpy(*tzn, tzname[tm->tm_isdst], MAXTZLEN + 1);
 				if (strlen(tzname[tm->tm_isdst]) > MAXTZLEN)
-					elog(WARNING, "Invalid timezone \'%s\'", tzname[tm->tm_isdst]);
+					ereport(WARNING,
+							(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+							 errmsg("invalid time zone name: \"%s\"",
+									tzname[tm->tm_isdst])));
 			}
 		}
 	}
 	else
 		tm->tm_isdst = -1;
-#endif
 #else							/* not (HAVE_TM_ZONE || HAVE_INT_TIMEZONE) */
 	if (tzp != NULL)
 	{
@@ -391,26 +299,16 @@ abstime2tm(AbsoluteTime _time, int *tzp, struct tm * tm, char **tzn)
 		}
 		else
 		{
-			*tzp = tb.timezone * 60;
-
-			/*
-			 * XXX does this work to get the local timezone string in V7?
-			 * - tgl 97/03/18
-			 */
+			/* default to UTC */
+			*tzp = 0;
 			if (tzn != NULL)
-			{
-				strftime(*tzn, MAXTZLEN, "%Z", localtime(&now));
-				tzn[MAXTZLEN] = '\0';	/* let's just be sure it's
-										 * null-terminated */
-			}
+				*tzn = NULL;
 		}
 	}
 	else
 		tm->tm_isdst = -1;
 #endif
-
-	return;
-}	/* abstime2tm() */
+}
 
 
 /* tm2abstime()
@@ -432,7 +330,7 @@ tm2abstime(struct tm * tm, int tz)
 		|| tm->tm_sec < 0 || tm->tm_sec > 60)
 		return INVALID_ABSTIME;
 
-	day = (date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(1970, 1, 1));
+	day = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - UNIX_EPOCH_JDATE;
 
 	/* check for time out of range */
 	if ((day < MIN_DAYNUM) || (day > MAX_DAYNUM))
@@ -451,14 +349,14 @@ tm2abstime(struct tm * tm, int tz)
 		return INVALID_ABSTIME;
 
 	return sec;
-}	/* tm2abstime() */
+}
 
 
-/* nabstimein()
+/* abstimein()
  * Decode date/time string and return abstime.
  */
 Datum
-nabstimein(PG_FUNCTION_ARGS)
+abstimein(PG_FUNCTION_ARGS)
 {
 	char	   *str = PG_GETARG_CSTRING(0);
 	AbsoluteTime result;
@@ -466,18 +364,19 @@ nabstimein(PG_FUNCTION_ARGS)
 	int			tz = 0;
 	struct tm	date,
 			   *tm = &date;
+	int			dterr;
 	char	   *field[MAXDATEFIELDS];
-	char		lowstr[MAXDATELEN + 1];
+	char		workbuf[MAXDATELEN + 1];
 	int			dtype;
 	int			nf,
 				ftype[MAXDATEFIELDS];
 
-	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad abstime external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	  || (DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tz) != 0))
-		elog(ERROR, "Bad abstime external representation '%s'", str);
+	dterr = ParseDateTime(str, workbuf, sizeof(workbuf),
+						  field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tz);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "abstime");
 
 	switch (dtype)
 	{
@@ -507,7 +406,8 @@ nabstimein(PG_FUNCTION_ARGS)
 			break;
 
 		default:
-			elog(ERROR, "Bad abstime (internal coding error) '%s'", str);
+			elog(ERROR, "unexpected dtype %d while parsing abstime \"%s\"",
+				 dtype, str);
 			result = INVALID_ABSTIME;
 			break;
 	};
@@ -516,11 +416,11 @@ nabstimein(PG_FUNCTION_ARGS)
 }
 
 
-/* nabstimeout()
+/* abstimeout()
  * Given an AbsoluteTime return the English text version of the date
  */
 Datum
-nabstimeout(PG_FUNCTION_ARGS)
+abstimeout(PG_FUNCTION_ARGS)
 {
 	AbsoluteTime time = PG_GETARG_ABSOLUTETIME(0);
 	char	   *result;
@@ -557,6 +457,31 @@ nabstimeout(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		abstimerecv			- converts external binary format to abstime
+ */
+Datum
+abstimerecv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+	PG_RETURN_ABSOLUTETIME((AbsoluteTime) pq_getmsgint(buf, sizeof(AbsoluteTime)));
+}
+
+/*
+ *		abstimesend			- converts abstime to binary format
+ */
+Datum
+abstimesend(PG_FUNCTION_ARGS)
+{
+	AbsoluteTime time = PG_GETARG_ABSOLUTETIME(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, time, sizeof(time));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 
 /* abstime_finite()
  */
@@ -577,11 +502,11 @@ abstime_finite(PG_FUNCTION_ARGS)
 static int
 abstime_cmp_internal(AbsoluteTime a, AbsoluteTime b)
 {
-/*
- * We consider all INVALIDs to be equal and larger than any non-INVALID.
- * This is somewhat arbitrary; the important thing is to have a
- * consistent sort order.
- */
+	/*
+	 * We consider all INVALIDs to be equal and larger than any non-INVALID.
+	 * This is somewhat arbitrary; the important thing is to have a
+	 * consistent sort order.
+	 */
 	if (a == INVALID_ABSTIME)
 	{
 		if (b == INVALID_ABSTIME)
@@ -594,7 +519,7 @@ abstime_cmp_internal(AbsoluteTime a, AbsoluteTime b)
 		return -1;				/* non-INVALID < INVALID */
 
 #if 0
-/* CURRENT is no longer stored internally... */
+	/* CURRENT is no longer stored internally... */
 	/* XXX this is broken, should go away: */
 	if (a == CURRENT_ABSTIME)
 		a = GetCurrentTransactionStartTime();
@@ -698,7 +623,9 @@ timestamp_abstime(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		elog(ERROR, "Unable to convert timestamp to abstime");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 		result = INVALID_ABSTIME;
 	}
 
@@ -722,7 +649,9 @@ abstime_timestamp(PG_FUNCTION_ARGS)
 	switch (abstime)
 	{
 		case INVALID_ABSTIME:
-			elog(ERROR, "Unable to convert abstime 'invalid' to timestamp");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("cannot convert abstime \"invalid\" to timestamp")));
 			TIMESTAMP_NOBEGIN(result);
 			break;
 
@@ -737,8 +666,9 @@ abstime_timestamp(PG_FUNCTION_ARGS)
 		default:
 			abstime2tm(abstime, &tz, tm, &tzn);
 			if (tm2timestamp(tm, 0, NULL, &result) != 0)
-				elog(ERROR, "Unable to convert ABSTIME to TIMESTAMP"
-					 "\n\tabstime_timestamp() internal error");
+				ereport(ERROR,
+						(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+						 errmsg("timestamp out of range")));
 			break;
 	};
 
@@ -766,7 +696,9 @@ timestamptz_abstime(PG_FUNCTION_ARGS)
 		result = tm2abstime(tm, 0);
 	else
 	{
-		elog(ERROR, "Unable to convert timestamp to abstime");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 		result = INVALID_ABSTIME;
 	}
 
@@ -790,7 +722,9 @@ abstime_timestamptz(PG_FUNCTION_ARGS)
 	switch (abstime)
 	{
 		case INVALID_ABSTIME:
-			elog(ERROR, "Unable to convert abstime 'invalid' to timestamptz");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("cannot convert abstime \"invalid\" to timestamp")));
 			TIMESTAMP_NOBEGIN(result);
 			break;
 
@@ -805,8 +739,9 @@ abstime_timestamptz(PG_FUNCTION_ARGS)
 		default:
 			abstime2tm(abstime, &tz, tm, &tzn);
 			if (tm2timestamp(tm, 0, &tz, &result) != 0)
-				elog(ERROR, "Unable to convert ABSTIME to TIMESTAMP WITH TIME ZONE"
-					 "\n\tabstime_timestamptz() internal error");
+				ereport(ERROR,
+						(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+						 errmsg("timestamp out of range")));
 			break;
 	};
 
@@ -830,17 +765,22 @@ reltimein(PG_FUNCTION_ARGS)
 			   *tm = &tt;
 	fsec_t		fsec;
 	int			dtype;
+	int			dterr;
 	char	   *field[MAXDATEFIELDS];
 	int			nf,
 				ftype[MAXDATEFIELDS];
-	char		lowstr[MAXDATELEN + 1];
+	char		workbuf[MAXDATELEN + 1];
 
-	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad reltime external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-		|| (DecodeInterval(field, ftype, nf, &dtype, tm, &fsec) != 0))
-		elog(ERROR, "Bad reltime external representation '%s'", str);
+	dterr = ParseDateTime(str, workbuf, sizeof(workbuf),
+						  field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeInterval(field, ftype, nf, &dtype, tm, &fsec);
+	if (dterr != 0)
+	{
+		if (dterr == DTERR_FIELD_OVERFLOW)
+			dterr = DTERR_INTERVAL_OVERFLOW;
+		DateTimeParseError(dterr, str, "reltime");
+	}
 
 	switch (dtype)
 	{
@@ -850,7 +790,8 @@ reltimein(PG_FUNCTION_ARGS)
 			break;
 
 		default:
-			elog(ERROR, "Bad reltime (internal coding error) '%s'", str);
+			elog(ERROR, "unexpected dtype %d while parsing reltime \"%s\"",
+				 dtype, str);
 			result = INVALID_RELTIME;
 			break;
 	}
@@ -858,7 +799,6 @@ reltimein(PG_FUNCTION_ARGS)
 	PG_RETURN_RELATIVETIME(result);
 }
 
-
 /*
  *		reltimeout		- converts the internal format to a reltime string
  */
@@ -878,19 +818,44 @@ reltimeout(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		reltimerecv			- converts external binary format to reltime
+ */
+Datum
+reltimerecv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+	PG_RETURN_RELATIVETIME((RelativeTime) pq_getmsgint(buf, sizeof(RelativeTime)));
+}
+
+/*
+ *		reltimesend			- converts reltime to binary format
+ */
+Datum
+reltimesend(PG_FUNCTION_ARGS)
+{
+	RelativeTime time = PG_GETARG_RELATIVETIME(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, time, sizeof(time));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 
 static void
 reltime2tm(RelativeTime time, struct tm * tm)
 {
-	TMODULO(time, tm->tm_year, 31557600);
-	TMODULO(time, tm->tm_mon, 2592000);
-	TMODULO(time, tm->tm_mday, 86400);
-	TMODULO(time, tm->tm_hour, 3600);
-	TMODULO(time, tm->tm_min, 60);
-	TMODULO(time, tm->tm_sec, 1);
+	double		dtime = time;
 
-	return;
-}	/* reltime2tm() */
+	FMODULO(dtime, tm->tm_year, 31557600);
+	FMODULO(dtime, tm->tm_mon, 2592000);
+	FMODULO(dtime, tm->tm_mday, 86400);
+	FMODULO(dtime, tm->tm_hour, 3600);
+	FMODULO(dtime, tm->tm_min, 60);
+	FMODULO(dtime, tm->tm_sec, 1);
+}
 
 
 /*
@@ -908,7 +873,10 @@ tintervalin(PG_FUNCTION_ARGS)
 
 	interval = (TimeInterval) palloc(sizeof(TimeIntervalData));
 	if (istinterval(intervalstr, &t1, &t2) == 0)
-		elog(ERROR, "Unable to decode tinterval '%s'", intervalstr);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type tinterval: \"%s\"",
+						intervalstr)));
 
 	if (t1 == INVALID_ABSTIME || t2 == INVALID_ABSTIME)
 		interval->status = T_INTERVAL_INVAL;	/* undefined  */
@@ -926,7 +894,6 @@ tintervalin(PG_FUNCTION_ARGS)
 
 /*
  *		tintervalout	- converts an internal interval format to a string
- *
  */
 Datum
 tintervalout(PG_FUNCTION_ARGS)
@@ -935,26 +902,66 @@ tintervalout(PG_FUNCTION_ARGS)
 	char	   *i_str,
 			   *p;
 
-	i_str = (char *) palloc(T_INTERVAL_LEN);	/* ['...' '...'] */
+	i_str = (char *) palloc(T_INTERVAL_LEN);	/* ["..." "..."] */
 	strcpy(i_str, "[\"");
 	if (interval->status == T_INTERVAL_INVAL)
 		strcat(i_str, INVALID_INTERVAL_STR);
 	else
 	{
-		p = DatumGetCString(DirectFunctionCall1(nabstimeout,
+		p = DatumGetCString(DirectFunctionCall1(abstimeout,
 							   AbsoluteTimeGetDatum(interval->data[0])));
 		strcat(i_str, p);
 		pfree(p);
 		strcat(i_str, "\" \"");
-		p = DatumGetCString(DirectFunctionCall1(nabstimeout,
+		p = DatumGetCString(DirectFunctionCall1(abstimeout,
 							   AbsoluteTimeGetDatum(interval->data[1])));
 		strcat(i_str, p);
 		pfree(p);
 	}
-	strcat(i_str, "\"]\0");
+	strcat(i_str, "\"]");
 	PG_RETURN_CSTRING(i_str);
 }
 
+/*
+ *		tintervalrecv			- converts external binary format to tinterval
+ */
+Datum
+tintervalrecv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+	TimeInterval interval;
+
+	interval = (TimeInterval) palloc(sizeof(TimeIntervalData));
+
+	interval->status = pq_getmsgint(buf, sizeof(interval->status));
+	if (!(interval->status == T_INTERVAL_INVAL ||
+		  interval->status == T_INTERVAL_VALID))
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_BINARY_REPRESENTATION),
+				 errmsg("invalid status in external \"tinterval\" value")));
+
+	interval->data[0] = pq_getmsgint(buf, sizeof(interval->data[0]));
+	interval->data[1] = pq_getmsgint(buf, sizeof(interval->data[1]));
+
+	PG_RETURN_TIMEINTERVAL(interval);
+}
+
+/*
+ *		tintervalsend			- converts tinterval to binary format
+ */
+Datum
+tintervalsend(PG_FUNCTION_ARGS)
+{
+	TimeInterval interval = PG_GETARG_TIMEINTERVAL(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, interval->status, sizeof(interval->status));
+	pq_sendint(&buf, interval->data[0], sizeof(interval->data[0]));
+	pq_sendint(&buf, interval->data[1], sizeof(interval->data[1]));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 
 /*****************************************************************************
  *	 PUBLIC ROUTINES														 *
@@ -1021,7 +1028,9 @@ reltime_interval(PG_FUNCTION_ARGS)
 	switch (reltime)
 	{
 		case INVALID_RELTIME:
-			elog(ERROR, "Unable to convert reltime 'invalid' to interval");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			  errmsg("cannot convert reltime \"invalid\" to interval")));
 			result->time = 0;
 			result->month = 0;
 			break;
@@ -1179,22 +1188,42 @@ timenow(PG_FUNCTION_ARGS)
 }
 
 /*
- *		reltimeeq		- returns true iff arguments are equal
- *		reltimene		- returns true iff arguments are not equal
- *		reltimelt		- returns true iff t1 less than t2
- *		reltimegt		- returns true iff t1 greater than t2
- *		reltimele		- returns true iff t1 less than or equal to t2
- *		reltimege		- returns true iff t1 greater than or equal to t2
+ * reltime comparison routines
  */
+static int
+reltime_cmp_internal(RelativeTime a, RelativeTime b)
+{
+	/*
+	 * We consider all INVALIDs to be equal and larger than any non-INVALID.
+	 * This is somewhat arbitrary; the important thing is to have a
+	 * consistent sort order.
+	 */
+	if (a == INVALID_RELTIME)
+	{
+		if (b == INVALID_RELTIME)
+			return 0;			/* INVALID = INVALID */
+		else
+			return 1;			/* INVALID > non-INVALID */
+	}
+
+	if (b == INVALID_RELTIME)
+		return -1;				/* non-INVALID < INVALID */
+
+	if (a > b)
+		return 1;
+	else if (a == b)
+		return 0;
+	else
+		return -1;
+}
+
 Datum
 reltimeeq(PG_FUNCTION_ARGS)
 {
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 == t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) == 0);
 }
 
 Datum
@@ -1203,9 +1232,7 @@ reltimene(PG_FUNCTION_ARGS)
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 != t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) != 0);
 }
 
 Datum
@@ -1214,9 +1241,7 @@ reltimelt(PG_FUNCTION_ARGS)
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 < t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) < 0);
 }
 
 Datum
@@ -1225,9 +1250,7 @@ reltimegt(PG_FUNCTION_ARGS)
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 > t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) > 0);
 }
 
 Datum
@@ -1236,9 +1259,7 @@ reltimele(PG_FUNCTION_ARGS)
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 <= t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) <= 0);
 }
 
 Datum
@@ -1247,9 +1268,16 @@ reltimege(PG_FUNCTION_ARGS)
 	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
 	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
 
-	if (t1 == INVALID_RELTIME || t2 == INVALID_RELTIME)
-		PG_RETURN_BOOL(false);
-	PG_RETURN_BOOL(t1 >= t2);
+	PG_RETURN_BOOL(reltime_cmp_internal(t1, t2) >= 0);
+}
+
+Datum
+btreltimecmp(PG_FUNCTION_ARGS)
+{
+	RelativeTime t1 = PG_GETARG_RELATIVETIME(0);
+	RelativeTime t2 = PG_GETARG_RELATIVETIME(1);
+
+	PG_RETURN_INT32(reltime_cmp_internal(t1, t2));
 }
 
 
@@ -1276,34 +1304,62 @@ tintervalsame(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(false);
 }
 
-
 /*
- *		tintervaleq		- returns true iff interval i1 is equal to interval i2
- *		Check length of intervals.
+ * tinterval comparison routines
+ *
+ * Note: comparison is based on the lengths of the intervals, not on
+ * endpoint value.  This is pretty bogus, but since it's only a legacy
+ * datatype I'm not going to propose changing it.
  */
+static int
+tinterval_cmp_internal(TimeInterval a, TimeInterval b)
+{
+	bool		a_invalid;
+	bool		b_invalid;
+	AbsoluteTime a_len;
+	AbsoluteTime b_len;
+
+	/*
+	 * We consider all INVALIDs to be equal and larger than any non-INVALID.
+	 * This is somewhat arbitrary; the important thing is to have a
+	 * consistent sort order.
+	 */
+	a_invalid = ((a->status == T_INTERVAL_INVAL) ||
+				 (a->data[0] == INVALID_ABSTIME) ||
+				 (a->data[1] == INVALID_ABSTIME));
+	b_invalid = ((b->status == T_INTERVAL_INVAL) ||
+				 (b->data[0] == INVALID_ABSTIME) ||
+				 (b->data[1] == INVALID_ABSTIME));
+
+	if (a_invalid)
+	{
+		if (b_invalid)
+			return 0;			/* INVALID = INVALID */
+		else
+			return 1;			/* INVALID > non-INVALID */
+	}
+
+	if (b_invalid)
+		return -1;				/* non-INVALID < INVALID */
+
+	a_len = a->data[1] - a->data[0];
+	b_len = b->data[1] - b->data[0];
+
+	if (a_len > b_len)
+		return 1;
+	else if (a_len == b_len)
+		return 0;
+	else
+		return -1;
+}
+
 Datum
 tintervaleq(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
-
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
-
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) == (t21 - t20));
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) == 0);
 }
 
 Datum
@@ -1311,24 +1367,8 @@ tintervalne(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
-
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
-
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) != (t21 - t20));
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) != 0);
 }
 
 Datum
@@ -1336,24 +1376,8 @@ tintervallt(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
-
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
-
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) < (t21 - t20));
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) < 0);
 }
 
 Datum
@@ -1361,24 +1385,8 @@ tintervalle(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
-
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
-
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) <= (t21 - t20));
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) <= 0);
 }
 
 Datum
@@ -1386,24 +1394,8 @@ tintervalgt(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
-
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
-
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) > (t21 - t20));
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) > 0);
 }
 
 Datum
@@ -1411,24 +1403,17 @@ tintervalge(PG_FUNCTION_ARGS)
 {
 	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
 	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
-	AbsoluteTime t10,
-				t11,
-				t20,
-				t21;
 
-	if (i1->status == T_INTERVAL_INVAL || i2->status == T_INTERVAL_INVAL)
-		PG_RETURN_BOOL(false);
+	PG_RETURN_BOOL(tinterval_cmp_internal(i1, i2) >= 0);
+}
 
-	t10 = i1->data[0];
-	t11 = i1->data[1];
-	t20 = i2->data[0];
-	t21 = i2->data[1];
+Datum
+bttintervalcmp(PG_FUNCTION_ARGS)
+{
+	TimeInterval i1 = PG_GETARG_TIMEINTERVAL(0);
+	TimeInterval i2 = PG_GETARG_TIMEINTERVAL(1);
 
-	if ((t10 == INVALID_ABSTIME) || (t11 == INVALID_ABSTIME)
-		|| (t20 == INVALID_ABSTIME) || (t21 == INVALID_ABSTIME))
-		PG_RETURN_BOOL(false);
-
-	PG_RETURN_BOOL((t11 - t10) >= (t21 - t20));
+	PG_RETURN_INT32(tinterval_cmp_internal(i1, i2));
 }
 
 
@@ -1641,7 +1626,7 @@ istinterval(char *i_string,
 			break;
 	}
 	p++;
-	/* skip leading blanks up to "'" */
+	/* skip leading blanks up to '"' */
 	while ((c = *p) != '\0')
 	{
 		if (IsSpace(c))
@@ -1667,12 +1652,12 @@ istinterval(char *i_string,
 		p1++;
 	}
 	/* get the first date */
-	*i_start = DatumGetAbsoluteTime(DirectFunctionCall1(nabstimein,
+	*i_start = DatumGetAbsoluteTime(DirectFunctionCall1(abstimein,
 													CStringGetDatum(p)));
-	/* rechange NULL at the end of the first date to a "'" */
+	/* rechange NULL at the end of the first date to a '"' */
 	*p1 = '"';
 	p = ++p1;
-	/* skip blanks up to "'", beginning of second date */
+	/* skip blanks up to '"', beginning of second date */
 	while ((c = *p) != '\0')
 	{
 		if (IsSpace(c))
@@ -1695,9 +1680,9 @@ istinterval(char *i_string,
 		p1++;
 	}
 	/* get the second date */
-	*i_end = DatumGetAbsoluteTime(DirectFunctionCall1(nabstimein,
+	*i_end = DatumGetAbsoluteTime(DirectFunctionCall1(abstimein,
 													CStringGetDatum(p)));
-	/* rechange NULL at the end of the first date to a ''' */
+	/* rechange NULL at the end of the first date to a '"' */
 	*p1 = '"';
 	p = ++p1;
 	/* skip blanks up to ']' */
@@ -1739,10 +1724,12 @@ timeofday(PG_FUNCTION_ARGS)
 	char		buf[100];
 	text	   *result;
 	int			len;
+	time_t		tt;
 
 	gettimeofday(&tp, &tpz);
+	tt = (time_t) tp.tv_sec;
 	strftime(templ, sizeof(templ), "%a %b %d %H:%M:%S.%%06d %Y %Z",
-			 localtime((time_t *) &tp.tv_sec));
+			 localtime(&tt));
 	snprintf(buf, sizeof(buf), templ, tp.tv_usec);
 
 	len = VARHDRSZ + strlen(buf);
